# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Bugzilla HTTP REST API proxy.
#
# The Initial Developer of the Original Code is
# the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   The Original Software was written to the Glory of God by 
#   Gervase Markham <written.to.the.glory.of.God@gerv.net>.
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

package Bugzilla::API::Controller::Utils;

use strict;
use warnings;
use base 'Exporter';

use vars qw(@ISA @EXPORT @EXPORT_OK);

@EXPORT = qw(
  ok_or_error
  accepted_or_error
  created_or_error
  include_exclude_fields
  json_typeify
);

use Data::Dumper;

# Helpers for returning right result codes
sub ok_or_error {
  my ($ctrl, $c, $result) = @_;
  
  my $output_type = $c->req->preferred_content_type() || "application/json";
                    
  # Tell Perl what type each field is, so JSON serializes correctly
  if ($output_type eq "application/json" ||
      $output_type eq "application/javascript")
  {
    json_typeify(\$result);
  }
  
  if ($result->{'error'}) {
    error($ctrl, $c, $result);  
  }
  else {
    $ctrl->status_ok($c, entity => $result);  
  }
}

sub accepted_or_error {
  my ($ctrl, $c, $result) = @_;
  
  if ($result->{'error'}) {
    error($ctrl, $c, $result);  
  }
  else {
    $ctrl->status_accepted($c, entity => $result);  
  }
}

sub created_or_error {
  my ($ctrl, $c, $result) = @_;
  
  if ($result->{'error'}) {
    error($ctrl, $c, $result);  
  }
  else {
    $ctrl->status_created($c, 
                          location => $result->{'ref'}, 
                          entity => $result);  
  }
}

# Errors need an "error" member and a "message" member.
sub error {
  my ($ctrl, $c, $result) = @_;
  
  if (!$result->{'message'}) {
    $result->{'message'} = "XXX Need to add message for this error";
  }
  
  $c->log->error(Dumper($result));
  
  if ($result->{'http_code'} && $result->{'http_code'} == 404) {
    $ctrl->status_not_found($c, message => $result->{'message'});
  }
  else {
    bad_request($ctrl, $c, $result);
  }
}

# We can't use the Catalyst::Controller::REST helper because it doesn't allow
# us to return a full object, just a string.
sub bad_request {
  my ($ctrl, $c, $result) = @_;

  $c->response->status(400);
  $c->log->debug("Status Bad Request: " . $result->{'message'}) if $c->debug;
  $ctrl->_set_entity($c, $result);
}

# $result is both an in and out parameter - the thing it points to gets
# modified.
#
# This sub only works for places where _default == _all, or the fields which
# make up the difference are known, enumerated and passed in as an arrayref.
sub include_exclude_fields {
  my ($c, $result, $differences) = @_;
  
  # $result can be undef (e.g. if there was an error)
  return if !$result;
  
  # Upscale non-array to array to make processing more uniform
  if (!(ref($result) eq "ARRAY")) {
    $result = [$result];
  }
  
  my @include_fields = split(/\s*,\s*/, 
                             $c->req->param('include_fields') || "");
  my %include_fields = map { $_ => 1 } @include_fields;
  my @exclude_fields = split(/\s*,\s*/, 
                             $c->req->param('exclude_fields') || "");
  
  # For the moment, we assume _default == _all...
  my $all = !@include_fields || 
            $include_fields{"_all"} || 
            $include_fields{"_default"} ||
            0;
  
  # But now, if we are _defaulting, and _default != _all, we need to 
  # _exclude_ any field in the set which makes the difference between 
  # _default and _all and which is not explicitly _included_.
  if ($include_fields{"_default"} && $differences) {
    foreach my $field (@$differences) {
      if (!$include_fields{$field}) {
        push(@exclude_fields, $field);
      }
    }
  }

  foreach my $obj (@$result) {
    if (!$all) {
      # Remove everything that's not an include_field
      foreach my $key (keys %$obj) {
        if (!$include_fields{$key}) {
          delete($obj->{$key});
        }
      }
    }
    
    # Remove exclude_fields
    foreach my $exclude (@exclude_fields) {
      delete($obj->{$exclude});
    }
  }
}

##############################################################################
# Typeification
##############################################################################
sub numify {
  my $scalarref = shift;
  
  if (ref($scalarref) eq "SCALAR") {
    $$scalarref += 0;
    return 1;
  }
  
  return 0;
}

# Note: sometimes can be called on non-array if there is a field name clash
sub numify_array {
  my $arrayrefref = shift;
  if (ref($$arrayrefref) eq "ARRAY") 
  {
    foreach my $item (@{$$arrayrefref}) {
      $item += 0;
    }
    
    return 1;
  }
  
  return 0;
}

sub json_boolify {
  my $scalarref = shift;
  if (ref($scalarref) eq "SCALAR") {
    $$scalarref = $$scalarref ? JSON::true : JSON::false;
    return 1;
  }
  
  return 0;
}

our $typefns = {
  # Bugs
  'actual_time'             => \&numify,
  'blocks'                  => \&numify_array,
  'depends_on'              => \&numify_array,
  'dupe_of'                 => \&numify,
  'estimated_time'          => \&numify,
  'id'                      => \&numify,
  'remaining_time'          => \&numify,
  'work_time'               => \&numify,
  'is_cc_accessible'        => \&json_boolify,
  'is_confirmed'            => \&json_boolify,
  'is_creator_accessible'   => \&json_boolify,
  
  # Other things
  'type_id'                     => \&numify,
  'bug_id'                      => \&numify,
  'size'                        => \&numify,
  'code'                        => \&numify,
  'http_code'                   => \&numify,
  'xmlrpc_code'                 => \&numify,
  'max_attachment_size'         => \&numify,
  'flag_type'                   => \&numify_array,
  'type'                        => \&numify,
  'is_private'                  => \&json_boolify,
  'is_obsolete'                 => \&json_boolify,
  'is_patch'                    => \&json_boolify,
  'error'                       => \&json_boolify,
  'is_accepting_bugs'           => \&json_boolify,
  'is_active'                   => \&json_boolify,
  'is_for_bugs'                 => \&json_boolify,
  'is_requestable'              => \&json_boolify,
  'is_multiplicable'            => \&json_boolify,
  'is_specifically_requestable' => \&json_boolify,
  'is_on_bug_entry'             => \&json_boolify,
  'email_enabled'               => \&json_boolify,
  'can_login'                   => \&json_boolify,
};

# This function coerces Perl scalars to the appropriate type so the JSON 
# serializer serializes them correctly.
#
# Note: parameter is reference to thing to typeify, even if thing it itself
# a reference to other things.
sub json_typeify {
  my ($thingref, $field) = @_;
  my $thing = $$thingref;
  my $ref = ref($thing);
  
  # Do we have a function for this specific field?
  if ($field && $typefns->{$field}) {
    my $result = &{$typefns->{$field}}($thingref);
    if ($result) {
      # No more typeification necessary
      return;
    }
  }
  
  if ($ref eq "HASH") {
    foreach my $key (keys %$thing) {
      json_typeify(\$thing->{$key}, $key);
    }
  }
  elsif ($ref eq "ARRAY") {
    foreach my $member (@$thing) {
      # Array members retain the field name of the array itself
      json_typeify(\$member, $field);
    }
  }
}

1;
