# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Bugzilla HTTP REST API proxy.
#
# The Initial Developer of the Original Code is
# the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   The Original Software was written to the Glory of God by 
#   Gervase Markham <written.to.the.glory.of.God@gerv.net>.
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK ****

package Bugzilla::API::Model::Attachment;

use strict;
use warnings;
use parent 'Catalyst::Model';

use LWP::UserAgent;
use HTTP::Request;
use HTTP::Request::Common;
use URI;
use URI::QueryParam;
use Data::Dumper;
use MIME::Base64;
use Bugzilla::API::Model::Utils;
use Bugzilla::API::Request;

##############################################################################
# Get one attachment
##############################################################################
# This is fairly horrible. We want the interface to require only the attachment
# ID, but the only way we have of getting the attachment data which doesn't
# involve scraping the HTML page is by getting it from the bug XML. And for
# that, we need the bug ID, which we don't have. So we have to scrape the
# page for that first.

# Cache of attachment to bug mappings, to make this slightly less painful
my %bug_for;

sub get {
  my ($self, $c, $attachid) = @_;

  my $result;
    
  if (!$bug_for{$attachid}) {
    my $url = new URI($c->config->{'bugzilla_url'} . 'attachment.cgi');
    $url->query_param('id', $attachid);
    $url->query_param('action', 'edit');
    
    my $res = make_request($c, "GET", $url);
    
    # Parse out bug ID
    if ($res->is_success && 
        $res->decoded_content =~ /show_bug.cgi\?id=(\d+)/) 
    {
      $bug_for{$attachid} = $1;
    }
    else {
      $result = http_result2error($res);
    }
  }
  
  my $include_fields = get_include_fields($c);

  # Get bug with attachment data
  if ($bug_for{$attachid}) {
    my $query = $c->req->params;

    $query->{'attachmentdata'} = ($include_fields->{"_all"} || 
                                  $include_fields->{"data"} ||
                                  $include_fields->{"encoding"} || 
                                  $c->req->params->{'attachmentdata'} 
                                  ? 1 : 0); 
    
    $result = Bugzilla::API::Model::Bug->get($c, $bug_for{$attachid}, $query);
    
    if (!$result->{'error'} && $result->{'attachments'}) {
      foreach my $attachment (@{$result->{'attachments'}}) {
        if ($attachment->{'id'} == $attachid) {
          $result = $attachment;
          last;
        }
      }
    }
  }
  
  if (!$result || !$result->{'attacher'}) {
    $result = { error => 1, 
                http_code => 404,
                message => "Attachment not found" };
  }
  
  return $result;
}

##############################################################################
# Update attachment metadata
##############################################################################
sub update {
  my ($self, $c, $attachid, $attachment) = @_;
  
  my $url = new URI($c->config->{'bugzilla_url'} . 'attachment.cgi');

  attachment2url($c, $attachment, $url);
  
  $url->query_param('id', $attachid);
  $url->query_param('action', 'update');

  my $res = make_request($c, "POST", $url);
  
  my $result = { "ok" => 1 };

  if ($res->decoded_content !~ /Changes Submitted to Attachment/) {
    $result = http_result2error($res);
  }
    
  return($result);
}

##############################################################################
# Create new attachment
##############################################################################
sub create {
  my ($self, $c, $bugid, $attachment) = @_;
  
  # We use a URI to collect parameters because attachment2url works that way
  # and we can reuse that code.
  my $tempurl = new URI();
  attachment2url($c, $attachment, $tempurl);
  
  $tempurl->query_param('bugid', $bugid);
  $tempurl->query_param('action', 'insert');
    
  # Get parameters back out...
  my $content = $tempurl->query_form_hash();
  
  # base64-encoded, we hope
  if ($attachment->{'encoding'} ne "base64") {
    return { 'error' => 1, 'message' => 'Unsupported encoding' }; 
  }
  
  my $data = decode_base64($attachment->{'data'}); 
  $content->{'data'} = [undef, 
                        $content->{'filename'}, 
                        "Content-Type" => $content->{'contenttypeentry'},
                        Content => $data];

  my $url = new URI($c->config->{'bugzilla_url'} . 'attachment.cgi');
  
  my $res = submit_form($c, $url, $content);
  
  my $result;
  
  if ($res->decoded_content =~ /Attachment (\d+) added to Bug $bugid/) {
    my $attachid = $1;
    my $ref = $c->uri_for("/attachment", $attachid)->as_string;      
    $result = { 'ref' => $ref, 'id' => $attachid };
  }
  else {
    $result = http_result2error($res);
  }
  
  return($result);  
}

##############################################################################
##############################################################################

# Change input field names into the ones Bugzilla likes
sub attachment2url {
  my ($c, $attachment, $url) = @_;

  foreach my $field ('description', 'file_name', 'content_type', 'is_private',
                     'is_patch', 'is_obsolete', 'update_token') 
  {
    if ($attachment->{$field}) {
      $url->query_param($NEWATTACH2OLD->{$field} || $field
                                                     => $attachment->{$field});
      delete($attachment->{$field});
    }
  }

  if (exists($attachment->{'comments'}->[0]->{'text'})) {
    $url->query_param("comment", $attachment->{'comments'}->[0]->{'text'});
  }
  
  # Outside if() test because of autovivification done by exists().
  delete($attachment->{'comments'});
  
  if ($attachment->{'last_change_time'}) {
    unfix_time($c, $attachment, 'last_change_time');
    $url->query_param("delta_ts", delete($attachment->{'last_change_time'}));
  }

  # Semi-temporary; would ideally convert this whole current function to be
  # not using URLs...
  my %params;
  flags2hash($attachment, \%params);
  
  foreach my $key (keys %params) {
    $url->query_param($key, $params{$key});
  }
  
  # content_type translates above to contenttypeentry
  if ($url->query_param('contenttypeentry')) {
    $url->query_param('contenttypemethod', 'manual');
  }
}

1;
