# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Bugzilla HTTP REST API proxy.
#
# The Initial Developer of the Original Code is
# the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   The Original Software was written to the Glory of God by 
#   Gervase Markham <written.to.the.glory.of.God@gerv.net>.
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

package Bugzilla::API::Model::Bug;

use strict;
use warnings;
use parent 'Catalyst::Model';

use Text::CSV_XS;
use LWP::UserAgent;
use URI::Escape;
use XML::Simple;
use Data::Dumper;
use BZ::Client;
use Bugzilla::API::Request;
use URI;
use URI::QueryParam;
use Array::Diff;
use Bugzilla::API::Model::Utils;

##############################################################################
# Bug list
##############################################################################
sub list {
  my ($self, $c, $query) = @_;

  my $url = new URI($c->config->{'bugzilla_url'} . 'buglist.cgi');
  $url->query_form($query);
  
  downgrade_search_url($c, $url);

  $url->query_param('columnlist', "all");
  $url->query_param('ctype', "csv");  
  
  # get default fields (ordinary search)
  my $res = make_request($c, "GET", $url);

  my $result;
  my @ids;

  # Check the outcome
  if ($res->is_success && $res->decoded_content =~ /^bug_id/s) {
    $result = { 'bugs' => [] };    

    # Parse CSV returned into data structure
    my $csv = Text::CSV_XS->new ({ binary => 1 });

    # Convert string $res->decoded_content to filehandle $fh
    open(my $fh, "<:utf8", \$res->decoded_content);
    
    # Get column names from the returned file, not from the list above, 
    # because we aren't guaranteed to get everything we ask for.
    $csv->column_names($csv->getline($fh));
    
    while (my $hashref = $csv->getline_hr($fh)) {
      $hashref = fixup_csv_bug($c, $hashref);
      push(@{$result->{'bugs'}}, $hashref);
      push(@ids, $hashref->{'id'});
    }    
  }
  else {
    $result = http_result2error($res);
  }

  ############################################################################
  # The include_fields algorithm is complex because we have to deal with 
  # _all and _default, and I want to avoid having to hard-code the list of
  # fields returned by default.
  # 
  # It works as follows:
  #  
  # get default fields (ordinary search)
  # 
  # if (include_fields) {
  #   if (_all) { 
  #     get full data by ID using /bug
  #     replace search data structure with this one
  #   }
  #   else {
  #     check include_fields against data
  #     if any fields missing { 
  #       get full data by ID using /bug
  #       if (_default) {
  #         Copy specified additional fields to original data structure
  #       }
  #       else {
  #         replace search data structure with this one
  #       }
  #     }
  #     
  #     if (!default) {
  #       reduce structure to specified fields
  #     }
  #   }
  # }

  my $include_fields = get_include_fields($c);
  
  if ((defined($result->{'bugs'}->[0]) && keys %$include_fields)) {
    # Do include_fields
    if ($include_fields->{'_all'}) {
      # _all has been specified; grab everything and use that instead, no 
      # matter what other (redundant) fields have also been listed
      $result = $self->get($c, \@ids);
    }
    else {
      # See whether we need fields other than the ones we already have
      my @needed_fields;
      foreach my $include_field (keys %$include_fields) {
        next if $include_field eq "_default";
        
        if (!$result->{'bugs'}->[0]->{$include_field}) {
          push (@needed_fields, $include_field);
        }
      }
      
      if (@needed_fields) {
        # We need more fields. Get the full data.
        my $fullresult = $self->get($c, \@ids);
        if ($include_fields->{'_default'}) {
          # Fields are specified as "_default + some"; so simplest 
          # method of construction, which doesn't require knowing what the
          # default list is, is to copy additional fields from new 
          # full data into original data.
          my %full_by_id = map { $_->{'id'} => $_ } @{ $fullresult->{'bugs'} };
          
          foreach my $bug (@{ $result->{'bugs'} }) {
            foreach my $needed_field (@needed_fields) {
              $bug->{$needed_field} =            
                                $full_by_id{ $bug->{'id'} }->{$needed_field};
            }
          }
        }
        else {
          # Fields are specified as explicit set. Replace limited data 
          # with full data, which is cut down to required set just below.
          $result = $fullresult;
        }
      }
       
      if (!$include_fields->{'_default'}) {
        # Only explicit set was requested; cut down data to that set.
        foreach my $bug (@{ $result->{'bugs'} }) {
          foreach my $field (keys %$bug) {
            if (!$include_fields->{$field}) {
              delete($bug->{$field});
            }
          }
        }        
      }
    }
  }  
 
  # Do exclude_fields
  exclude_fields($c, $result->{'bugs'});
  
  return $result;
}

##############################################################################
# Get single bug
# $bugids can be a scalar or an arrayref containing scalars
# The return value of this function will be a hashref - either a single bug
# if you passed a scalar, or an error, or a { 'bugs' => [ {} {}... ] } hashref
# if you passed an arrayref.
##############################################################################
sub get {
  my ($self, $c, $bugids) = @_;

  my $url = new URI($c->config->{'bugzilla_url'} . 'show_bug.cgi');

  my $ids = ref($bugids) ? $bugids : [$bugids];
  $url->query_param('id', join(",", @$ids));
  
  $url->query_param('ctype', "xml");
  
  my $include_fields = get_include_fields($c);

  # We have a different default for attachment data and comments to the 
  # standard XML interface.  We also use saner field names on the API.
  if (!$include_fields->{'attachments.data'} &&
      !$include_fields->{'attachments.encoding'} &&
      !$include_fields->{'_all'} &&
      !$c->req->param("attachmentdata")) 
  {
    $url->query_param_append("excludefield" => "attachmentdata");
  }

  if (!$include_fields->{'comments'} &&
      !$include_fields->{'_all'}) 
  {
    $url->query_param_append("excludefield" => "long_desc");
  }
  
  my $res = make_request($c, "GET", $url);

  my $result = {};
  
  # Check the outcome
  if ($res->decoded_content =~ /^<\?xml/) {
    my $xs = XML::Simple->new(ForceArray => 
                                ['cc', 'see_also', 'attachment', 'depends_on', 
                                 'blocked', 'comment', 'flag', 'group', 'bug',
                                 qr/^cf_multi/],
                              KeyAttr => { },
                              SuppressEmpty => 1
                              );
    
    my $hashref = $xs->XMLin($res->decoded_content);
    
    if ($hashref) {
      my $bugs = {};
      
      foreach my $bug (@{ $hashref->{'bug'} }) {
        # This isn't checking for an HTTP API error, but the way the XML 
        # version of the bug returns an error.
        if (!$bug->{'error'}) {
          $bug = fixup_xml_bug($c, $bug);
          
          $bugs->{$bug->{'id'}} = $bug;
        }
        else {
          my $error = xml_error2error($bug);
          $bugs->{$bug->{'bug_id'}} = $error;
        }
      }

      # XML doesn't return history, but we offer the option of getting it
      # alongside the rest of the bug info. So we make another call.
      if ($include_fields->{'history'} ||
          $include_fields->{'_all'}) 
      {
        my $history = $self->history($c, $ids);
        if (!$history->{'error'}) {
          foreach my $hbug (@{ $history->{'bugs'} }) {
              $bugs->{$hbug->{'id'}}->{'history'} = $hbug->{'history'};
          }          
        }
        else {
            # Ignore history errors; easier than some kind of nested error
            # reporting.
        }
      }

      my @vals = values(%$bugs);
      $result = ref($bugids) ? { 'bugs' => \@vals } : $vals[0];
    }
    else {
      # XML parser error
      $result = {
        error => 1,
        message => 'Parse error in XML sent by Bugzilla (bad characters?)'
      };
    }    
  }
  else {
    $result = http_result2error($res);
  }  
  
  return $result;
}

##############################################################################
# Update single bug
##############################################################################
sub update {
  my ($self, $c, $bugid, $bug, $force) = @_;
  
  my $url = new URI($c->config->{'bugzilla_url'} . 'process_bug.cgi');

  my %params;

  remove_immutables($bug);

  # People
  my @people = ("assigned_to", "qa_contact"); # Not creator :-)  
  foreach my $field (@people) {
    if (exists($bug->{$field})) {
      if (defined($bug->{$field})) {
        if ($bug->{$field}->{'name'}) {
          $params{$field} = $bug->{$field}->{'name'};
        }
        else {
          $params{$field} = "";
        }
      }
      else {
        # value is special value 'undef' - this means 'reset to default'
        # <sigh> consistent field names would be nice.
        my %people_sd_map = ('assigned_to' => 'set_default_assignee',
                             'qa_contact' => 'set_default_qa_contact');
          $params{$people_sd_map{$field}} = 1;
      }
      
      delete($bug->{$field});    
    }
  }
  
  # API user is responsible for making sure they have picked valid component,
  # version and TM values; that includes on a product change. So we set this
  # always; if they've picked bad values, Bugzilla will throw an error.
  $params{'confirm_product_change'} = '1';

  # This value is used if there's a mid-air collision; it's the index of the
  # last comment. We can't calculate it accurately, because the API user is
  # under no obligation to supply all the bug's comments on the change 
  # request, and even if they did, that wouldn't necessarily be all of them
  # due to comment privacy. All it's used for is displaying the right comments
  # on the mid-air template; setting it to a high value means that this 
  # template will never display comments, but that's OK as we don't use the
  # data it supplies.
  $params{"longdesclength"} = "10000";

  if ($bug->{'last_change_time'}) {
    unfix_time($c, $bug, 'last_change_time');
    $params{"delta_ts"} = delete($bug->{'last_change_time'});
  }
  
  # Comma-separated multi-values fields
  foreach my $field ('blocks', 'depends_on', 'keywords') {
    if ($bug->{$field}) {
      $params{new2old($field)} = join(", ", @{$bug->{$field}});
      delete($bug->{$field});      
    }
  }
  
  # Processing some fields requires a copy of current bug data
  if ($bug->{'cc'} ||
      $bug->{'see_also'} ||
      $bug->{'groups'} ||
      $force)
  {
    my $currentbug = $self->get($c, $bugid, $c->req->params);
    if ($currentbug->{'error'}) {
      $currentbug->{'message'} = "Error fetching copy of existing bug: " 
                                 . $currentbug->{'message'};
      return($currentbug);
    }

    # If the bug object specifies CCs, then we are making the list match    
    if ($bug->{'cc'}) {
      my $currentccs = extract_cc_names($currentbug);
      my $newccs = extract_cc_names($bug);
    
      $params{"removecc"} = 1;
      arraydiff2addremove($currentccs, $newccs, \%params, "newcc", "cc");
      delete($bug->{'cc'});
    }
    
    # Same for see_also
    if ($bug->{'see_also'}) {
      arraydiff2addremove($currentbug->{'see_also'}, $bug->{'see_also'}, 
                          \%params, "see_also", "remove_see_also");
      delete($bug->{'see_also'});
    }
    
    # Groups
    if ($bug->{'groups'}) {
      if ($c->config->{'bugzilla_version'} < 4.0) {
        # 3.6 and below
        # Define the bit as off for every old group...
        if ($currentbug->{'groups'}) {
          foreach my $group (@{ $currentbug->{'groups'} }) {
            my $gid = $group->{'id'};
            $params{"defined_bit-" . $gid} = 1;
            $params{"bit-" . $gid} = 0;
          }
        }
        
        # ...and redefine as on for every current group.
        foreach my $group (@{ $bug->{'groups'} }) {
          my $gid = $group->{'id'};
          $params{"defined_bit-" . $gid} = 1;
          $params{"bit-" . $gid} = 1;
        }
      }
      else {
        # 4.0 and above
        $params{"defined_groups"} = [];
        $params{"groups"} = [];
        
        # 'defined_groups' are the current ones...
        if ($currentbug->{'groups'}) {
          foreach my $group (@{ $currentbug->{'groups'} }) {
            my $gname = $group->{'name'};
            push(@{$params{"defined_groups"}}, $gname);
          }
        }
        
        # ...and 'groups' are the ones you want.
        foreach my $group (@{ $bug->{'groups'} }) {
          my $gname = $group->{'name'};
          push(@{$params{"groups"}}, $gname);
        }
      }
      
      delete($bug->{'groups'});
    }
 
    # "Force" is used when submitting an update directly after a creation;
    # This way we don't have to do an _extra_ bug request to get a token.
    if ($force) {
      $params{'token'} = $currentbug->{'update_token'};
    }
  }
  
  # Flags
  if ($bug->{'flags'}) {
    flags2hash($bug, \%params);
    delete($bug->{'flags'});
  }

  # Comments - add a new one if there's one without an ID, or tweak privacy
  if ($bug->{'comments'}) {
    foreach my $comment (@{ $bug->{'comments'} }) {
      if ($comment->{'id'}) {
        # Existing comment; tweak privacy flags if necessary
        my $id = $comment->{'id'};
        $params{"defined_isprivate_$id"} = 1;
        $params{"isprivate_$id"} = 
                                 ($comment->{'is_private'} && 
                                  $comment->{'is_private'} eq "true") ? 1 : 0;
      }
      else {
        # New comment to be added
        # If multiple new comments are specified, only the last one will be
        # added.
        my $comment_privacy_key = 
          $c->config->{'bugzilla_version'} >= 4.0 ? "comment_is_private"  
                                                  : "commentprivacy";
        $params{$comment_privacy_key} = 
                                 ($comment->{'is_private'} && 
                                  $comment->{'is_private'} eq "true") ? 1 : 0;
        $params{"comment"} = $comment->{'text'};
      }
    }
    
    delete($bug->{'comments'});
  }

  # Delete things we know we currently don't support on the bug update call
  delete($bug->{'attachments'});
  
  # Booleans
  foreach my $field ("is_cc_accessible", "is_creator_accessible",
                     "is_confirmed") 
  {
    if (defined($bug->{$field})) {
      $params{new2old($field)} = ($bug->{$field} eq "true") ? "1" : "0";
      delete($bug->{$field});
    }
  }

  # We use a generic mechanism for everything else; this includes custom
  # fields
  foreach my $field (keys %$bug) {
    # Either a multi-valued field in array ref or a single-valued field;
    # we do the same either way.
    $params{new2old($field)} = $bug->{$field};    
    delete($bug->{$field});
  }

  # We get this from the URL used, not from the bug data
  $params{'id'} = $bugid;

  my $res = make_request($c, "POST", $url, \%params);

  my $result = { "ok" => 1 };

  if ($res->decoded_content !~ /Changes submitted for/) {
    $result = http_result2error($res);
  }
  
  return $result;
}

##############################################################################
# Create new bug
##############################################################################
sub create {
  my ($self, $c, $bug) = @_;

  my $result;
  my $extra = {};
    
  # Downgrade user objects to email addresses
  foreach my $person ('assigned_to', 'creator', 'qa_contact') {
    if ($bug->{$person}) {
      $bug->{$person} = $bug->{$person}->{'name'};
    }
  }
  
  if ($bug->{'cc'}) {
    my @names = map ( { $_->{'name'} } @{$bug->{'cc'}});
    $bug->{'cc'} = \@names;
  }

  # For consistency, we take initial comment in comments array
  delete($bug->{'description'});
  if (ref($bug->{'comments'})) {
    $bug->{'description'} = $bug->{'comments'}->[0]->{'text'};
    delete($bug->{'comments'});
  }

  # Remove fields the XML-RPC interface will object to
  # We list legal fields rather than illegal ones because enumerating badness
  # breaks more easily. This list straight from the 3.4 documentation.
  my %legalfields = 
    map { $_ => 1 } qw(product component summary version description op_sys
                       platform priority severity alias assigned_to cc
                       qa_contact status target_milestone);

  foreach my $field (keys %$bug) {
    if (!$legalfields{$field} && $field !~ /^cf_/) {
      $extra->{$field} = $bug->{$field};
      delete($bug->{$field});
    }
  }

  $result = make_rpc_request($c, "Bug.create", $bug);
  
  if (!$result->{'error'}) {
    my $bugid = $result->{'id'};
    
    # So far so good. Do we need to slip in a bug update?
    remove_immutables($extra);
    # We shouldn't have one of these, but let's not mid-air if they send one
    delete($extra->{'last_change_time'});
    
    if (keys(%$extra)) {
      $result = $self->update($c, $bugid, $extra, "force");
    }
    
    if (!$result->{'error'}) {
      my $ref = $c->uri_for($c->action, $bugid)->as_string;      
      $result = { 
        'ref' => $ref,
        'id' => $bugid
      };
    }  
  }
  
  return $result;
}

##############################################################################
# Get history
# $bugids can be a scalar or an arrayref containing scalars
# The return value of this function will be a hashref with a single member,
# 'bugs', whose value is an arrayref of "bug" hashrefs, with "id", possibly 
# "alias", and "history" keys.
##############################################################################
sub history {
  my ($self, $c, $bugids) = @_;  

  my $result;
  
  my $ids = ref($bugids) ? $bugids : [$bugids];

  my $method = "Bug.history";
  my $payload = { "ids" => $ids };
  
  $result = make_rpc_request($c, "Bug.history", $payload);
    
  if (!$result->{'error'}) {
    foreach my $bug (@{ $result->{'bugs'} }) {
      foreach my $entry (@{ $bug->{'history'} }) {
        change_field_name($entry, 'who', 'changer');
        change_field_name($entry, 'when', 'change_time');
        fix_time($c, $entry, "change_time");
      
        foreach my $change (@{$entry->{'changes'}}) {
          # Upgrade field names
          my $field_name = $change->{'field_name'};
          $change->{'field_name'} = $OLD2NEW->{$field_name} || $field_name;
        }
      
        # Upgrade user
        fix_person($c, $entry, 'changer');
      }
    }
  }
  
  return $result;
}

##############################################################################
# Get comments
##############################################################################
sub comments {
  my ($self, $c, $bugid, $new_since) = @_;
  
  my $result = { error => 1 };
  
  if ($c->config->{'bugzilla_version'} >= 3.4) {
    # XML-RPC version, which supports new_since
    my $payload = { "ids" => [ $bugid ], "new_since" => $new_since };
    
    # 3.4 has trouble with correctly-formatted XML-RPC dates :-(
    # XXX But all this is broken, because we aren't sending it as a proper
    # date value in the XML-RPC, so it's bypassing the checking.
    if ($c->config->{'bugzilla_version'} >= 3.6) {
      # XML-RPC times are broken non-ISO forms
      if ($payload->{'new_since'}) {
        fix_time($c, $payload, 'new_since');
        $payload->{'new_since'} =~ s/-//g;
        $payload->{'new_since'} =~ s/Z$//;
      }
    }
    
    $result = make_rpc_request($c, "Bug.comments", $payload);
                  
    if (!$result->{'error'}) {
      $result = $result->{'bugs'}->{$bugid};
      
      fix_comments($c, $result); 
    }
  }
  else {
    # Get comments from legacy XML interface for older Bugzillas
    # No support for new_since (although it could be written)
    my $query = $c->req->params;
    $query->{'comments'} = 1;
    $result = $self->get($c, $bugid, $query);
    
    if (!$result->{'error'}) {
      if ($result->{'comments'}) {
        # Strip out other stuff
        $result = { 'comments' => $result->{'comments'} };
      }
      else {
        $result = { 'comments' => [] };
      }
    }
  }
  
  return $result;
}

##############################################################################
# Add comment
##############################################################################
sub add_comment {
  my ($self, $c, $bugid, $comment) = @_;
  
  my $result;
  
  my $private = ($comment->{'is_private'} && 
                ($comment->{'is_private'} eq "true")) ? 1 : 0;
                
  my $payload = { "id" => $bugid,
                  "comment" => $comment->{'text'},
                  "private" => $private
                };
  
  $result = make_rpc_request($c, "Bug.add_comment", $payload);
 
  if (!$result->{'error'}) {
    # Comments don't have their own refs, so pass ref of comment list
    my $ref = $c->uri_for("/bug", $bugid, "comment")->as_string;      
    $result = { 'ref' => $ref, 'id' => $result->{'id'} };
  }
 
  return $result;
}

##############################################################################
# Get attachments
##############################################################################
sub attachments {
  my ($self, $c, $bugid, $getdata) = @_;
  
  my $query = $c->req->params;

  my $include_fields = get_include_fields($c);

  $query->{'attachmentdata'} = ($include_fields->{"_all"} || 
                                $include_fields->{"data"} ||
                                $include_fields->{"encoding"} || 
                                $c->req->params->{'attachmentdata'} 
                                ? 1 : 0); 

  my $result = $self->get($c, $bugid, $query);
  
  if (!$result->{'error'}) {
    if ($result->{'attachments'}) {
      $result = { 'attachments' => $result->{'attachments'} };
    }
    else {
      $result = { 'attachments' => [] };
    }
  }
  
  return $result;
}

##############################################################################
# Get flags
##############################################################################
sub flags {
  my ($self, $c, $bugid) = @_;
  
  my $result = $self->get($c, $bugid, $c->req->params);
  
  if (!$result->{'error'}) {
    if ($result->{'flags'}) {
      $result = { 'flags' => $result->{'flags'} };
    }
    else {
      $result = { 'flags' => [] };
    }
  }
  
  return $result;
}

1;
